
package stamp.peripheral.sensor.GPS.Menu;

import stamp.core.*;
import stamp.peripheral.memory.eeprom.*;                // Contains handling for EEPROM operations
import stamp.peripheral.sensor.GPS.GPSMath.*;           // Contains the GPS calculations
import stamp.peripheral.sensor.GPS.NMEA.*;              // Contains the NMEA messages and other NMEA specific things
import stamp.peripheral.sensor.GPS.StringOperations.*;  // Contains some string handling classes
import stamp.peripheral.sensor.GPS.Include.*;           // Contains the linked list class and other stuff
import stamp.peripheral.sensor.GPS.Keyboard.*;          // Contains the memkey implementation
import stamp.peripheral.sensor.GPS.Temperature.*;       // Contains the SHT11 sensor wrapper
import stamp.peripheral.sensor.GPS.DisplayDriver.*;     // Contains the LCD driver for the CrystalFontz
import stamp.peripheral.sensor.GPS.RouteFollower.*;     // Contains the class handling keeping track of were we are and where to go

public class GPSMenuHandler
{
  public static final byte GPS_MENU_MAIN_MENU            = 0;
  public static final byte GPS_MENU_FOLLOW_ROUTE         = 1;
  public static final byte GPS_MENU_CREATE_ROUTE         = 2;
  public static final byte GPS_MENU_EDIT_ROUTE           = 3;
  public static final byte GPS_MENU_DELETE_ROUTE         = 4;
  public static final byte GPS_MENU_SET_HOME_POSITION    = 5;
  public static final byte GPS_MENU_SET_BACKLIGHT_MODE   = 6;
  public static final byte GPS_MENU_SET_LCD_CONTRAST     = 7;


  byte m_menuMode = GPS_MENU_MAIN_MENU;


  DisplayDriver m_displayDriver;       // Class used for handling the display
  GlobalVars m_globalVars;             // Contains global variables used in the whole program
  MemoryRouteHandler m_routeHandler;   // Memory handler for all the routes
  MEMKey m_keyboard;                   // Handles the hardware interface with the keyboeard
  RouteFollower m_routeFollower;       // The route follower is showing info about the current joined route
  GPSMath m_gpsMath;                   // Class handling all the math opearations
  NMEA0183 m_nmea0183;                 // Class keeping track of the current GPS messages
  Settings m_settings;                 // This class is storing/retrieving the settings from EEPROM memory

  public GPSMenuHandler( DisplayDriver displayDriver,
                         GlobalVars globalVars,
                         MemoryRouteHandler routeHandler,
                         MEMKey keyboard,
                         RouteFollower routeFollower,
                         GPSMath gpsMath,
                         NMEA0183 nmea0183,
                         Settings settings)
  {
    m_displayDriver = displayDriver;
    m_globalVars = globalVars;
    m_routeHandler = routeHandler;
    m_keyboard = keyboard;
    m_routeFollower = routeFollower;
    m_gpsMath = gpsMath;
    m_nmea0183 = nmea0183;
    m_settings = settings;
  }
  public void RunMenu()
  {
    m_menuMode = GPS_MENU_MAIN_MENU;
    m_displayDriver.ClearScreen(DisplayDriver.DISPLAY_1 | DisplayDriver.DISPLAY_2);

    m_displayDriver.WriteMessage(DisplayDriver.DISPLAY_1, 0, 0, "1: Follow route");
    m_displayDriver.WriteMessage(DisplayDriver.DISPLAY_1, 1, 0, "2: Create route");
    m_displayDriver.WriteMessage(DisplayDriver.DISPLAY_1, 2, 0, "3: Delete route");
    //m_displayDriver.WriteMessage(DisplayDriver.DISPLAY_1, 3, 0, "4: Set time offset");

    //m_displayDriver.WriteMessage(DisplayDriver.DISPLAY_2, 0, 0, "5: Backlight mode");

    m_displayDriver.WriteMessage(DisplayDriver.DISPLAY_2, 3, 0, "C=Cancel   #=Select");

    while (true)
    {
      // Expect the number...
      MEMKeyInput input = m_keyboard.GetChar( true,
                                              MEMKey.KEYBOARD_MODE_NUMERIC,
                                              m_displayDriver,
                                              DisplayDriver.DISPLAY_2,
                                              1,
                                              0,
                                              m_globalVars.m_buffer1,
                                              0,
                                              MEMKey.KEYBOARD_BYTE_C,
                                              MEMKey.KEYBOARD_BYTE_C,
                                              MEMKey.KEYBOARD_NONE_BYTE,
                                              MEMKey.KEYBOARD_NONE_BYTE,
                                              MEMKey.KEYBOARD_NONE_BYTE );

      if (input.m_Key == MEMKey.KEYBOARD_BYTE_C)
        break;

      m_displayDriver.ClearScreen(DisplayDriver.DISPLAY_1 | DisplayDriver.DISPLAY_2);
      m_displayDriver.WriteMessage(DisplayDriver.DISPLAY_2, 0, 0, "D=Yes");
      m_displayDriver.WriteMessage(DisplayDriver.DISPLAY_2, 0, 12, "C=Next");


      try
      {
        switch ( input.m_Numeric )
        {
          case 1: FollowRoute(); return;        // Follow route
          case 2: HandleCreateRoute(); return;  // Create route
          case 3: DeleteRoute(); return;        // Delete route
          //case 4: SetTimeOffset(); return;      // Time offset
          case 5: return;    // Backlight mode
        }
      }
      catch(Exception ex)
      {

      }
    }
  }
  private boolean AnyRoutes() throws Exception
  {
    // Nothing to do here
    int nNextRoute = m_routeHandler.GetNextAvailableRouteSlot();
    if (nNextRoute == 0)
    {
      m_displayDriver.WriteMessage(DisplayDriver.DISPLAY_1, 0, 0, "No stored routes...");
      CPU.delay(6000);
      return false;
    }
    return true;
  }
  /*
  private void SetTimeOffset()
  {
    m_displayDriver.WriteMessage(DisplayDriver.DISPLAY_1, 0, 0, "Set offset:");
    m_displayDriver.WriteMessage(DisplayDriver.DISPLAY_1, 0, 0, "Set offset:");

    m_displayDriver.WriteMessage(DisplayDriver.DISPLAY_1, 1, 0, "A=UTC+  B=UTC-");
    m_displayDriver.WriteMessage(DisplayDriver.DISPLAY_1, 3, 0, "C=Cancel");

    MEMKeyInput input = m_keyboard.GetChar( true,
                                            MEMKey.KEYBOARD_MODE_EXTRA, // "A", "B", "C", "D" and so on
                                            m_displayDriver,
                                            DisplayDriver.DISPLAY_1,
                                            0,
                                            0,
                                            m_globalVars.m_buffer1,
                                            0,
                                            MEMKey.KEYBOARD_BYTE_C,
                                            MEMKey.KEYBOARD_BYTE_C,
                                            MEMKey.KEYBOARD_NONE_BYTE,
                                            MEMKey.KEYBOARD_NONE_BYTE,
                                            MEMKey.KEYBOARD_NONE_BYTE );

    // Nothing to do here
    if (input.m_Key == MEMKey.KEYBOARD_BYTE_C)
      return;

    if (input.m_Key == MEMKey.KEYBOARD_BYTE_A) // UTC+
    {

    }
    else if (input.m_Key == MEMKey.KEYBOARD_BYTE_B) // UTC+
    {

    }


  }
  */
  private void FollowRoute() throws Exception
  {
    // Check if there are any routes
    if ( !AnyRoutes() ) return;
    m_displayDriver.WriteMessage(DisplayDriver.DISPLAY_1, 0, 0, "Follow route?");

    // For every route, let the user to choose the route he wants to join
    for ( int nRoute = 1; nRoute <= 36; nRoute++)
    {
      // Check if this is a route
      if (!m_routeHandler.IsRoute(nRoute)) continue;

      // If the user is choosing this route...
      if (ChooseRoute(nRoute))
      {
        // Check if there are any coordinates
        if (m_routeHandler.GetNumberOfCoordinatesInRoute(nRoute) == 0 )
        {
          m_displayDriver.WriteMessage(DisplayDriver.DISPLAY_1, 0, 0, "No stored coordinates...");
          CPU.delay(6000);
          return;
        }

        m_displayDriver.WriteMessage(DisplayDriver.DISPLAY_1, 2, 0, "Waypoint:" );

        // Let the user select the waypoint
        for (int nWaypoint = 1; nWaypoint < 41; nWaypoint++)
        {
          // Let this function handle the user
          int nWayPointSelected = ChooseWaypoint(nRoute, nWaypoint);

          // Special handling for moving forward to next waypoint
          if (nWayPointSelected == -1) continue;
          if ( nWayPointSelected == 0 ) return; // Nothing to do here (the user pushed Cancel button)

          {
            // This is the waypoint we selected
            int nMode = RouteFollower.FOLLOW_ASCENDING;

            m_displayDriver.ClearScreenRange(DisplayDriver.DISPLAY_1, 3, 4, 0, 20);
            m_displayDriver.ClearScreenRange(DisplayDriver.DISPLAY_2, 0, 3, 0, 20);
            m_displayDriver.WriteMessage(DisplayDriver.DISPLAY_2, 0, 0, "1: Ascending" );
            m_displayDriver.WriteMessage(DisplayDriver.DISPLAY_2, 1, 0, "2: Descending" );
            m_displayDriver.WriteMessage(DisplayDriver.DISPLAY_2, 3, 0, "C=Cancel");

            // Expect the number...
            MEMKeyInput input = m_keyboard.GetChar( true,
                                                    MEMKey.KEYBOARD_MODE_NUMERIC,
                                                    m_displayDriver,
                                                    DisplayDriver.DISPLAY_2,
                                                    1,
                                                    0,
                                                    m_globalVars.m_buffer1,
                                                    0,
                                                    MEMKey.KEYBOARD_BYTE_C,
                                                    MEMKey.KEYBOARD_BYTE_C,
                                                    MEMKey.KEYBOARD_NONE_BYTE,
                                                    MEMKey.KEYBOARD_NONE_BYTE,
                                                    MEMKey.KEYBOARD_NONE_BYTE );

            if (input.m_Key == MEMKey.KEYBOARD_BYTE_C)
              return;

            if (input.m_Numeric == 2 )
              nMode = RouteFollower.FOLLOW_DESCENDING;

            m_routeFollower.JoinRoute(nRoute, nWayPointSelected, nMode, m_displayDriver );
            return;
          }
        }

        m_displayDriver.ClearScreenRange(DisplayDriver.DISPLAY_1, 2, 3, 0, 20);
      }
    }

  }

  private int ChooseWaypoint(int nRoute, int nWaypoint) throws Exception
  {
    m_displayDriver.WriteMessage(DisplayDriver.DISPLAY_2, 1, 0, "A=Nearest");
    m_displayDriver.ClearScreenRange(DisplayDriver.DISPLAY_1, 2, 3, 11, 20);
    if ( !m_routeHandler.GetCoordinate( nRoute, nWaypoint, m_globalVars.m_gpsCoordinate1 ) ) return -1;

    m_globalVars.m_buffer1.clear();
    m_globalVars.m_buffer1.append(m_globalVars.m_gpsCoordinate1.GetCoordinateName().toString());
    m_displayDriver.WriteMessage(DisplayDriver.DISPLAY_1, 2, 11, m_globalVars.m_buffer1.toString() );

    // Expect the number...
    MEMKeyInput input = m_keyboard.GetChar( true,
                                            MEMKey.KEYBOARD_MODE_EXTRA,
                                            m_displayDriver,
                                            DisplayDriver.DISPLAY_1,
                                            0,
                                            0,
                                            m_globalVars.m_buffer1,
                                            0,
                                            MEMKey.KEYBOARD_BYTE_C,
                                            MEMKey.KEYBOARD_BYTE_D,
                                            MEMKey.KEYBOARD_NONE_BYTE,
                                            MEMKey.KEYBOARD_NONE_BYTE,
                                            MEMKey.KEYBOARD_NONE_BYTE );

      // This waypoint is chosen by the user
      if (input.m_Key == MEMKey.KEYBOARD_BYTE_D)
        return nWaypoint;

      // Let the user choose next waypoint
      if (input.m_Key == MEMKey.KEYBOARD_BYTE_C)
        return -1;

      // Find the nearest point...
      if (input.m_Key == MEMKey.KEYBOARD_BYTE_A)
      {
        GGA gga = m_nmea0183.GetGGAMessage();
        if (gga != null )
        {
          GPSCoordinate lastCoordinate = gga.GetGPSCoordinate();

          int nWayPointSelected = nWaypoint;
          m_gpsMath.CalculateDistance(lastCoordinate, m_globalVars.m_gpsCoordinate1, m_globalVars.m_fTemp12, GPSConvert.GPS_CONVERSION_NAUTICALMILES);
          m_displayDriver.ClearScreenRange(DisplayDriver.DISPLAY_2, 2, 3, 0, 20);
          m_displayDriver.WriteMessage(DisplayDriver.DISPLAY_2, 2, 0, "Calculating..." );

          // Now, for all the points in the list...find the closest one
          for ( int n = 1; n < 41; n++ )
          {
            if ( !m_routeHandler.GetCoordinate( nRoute, n, m_globalVars.m_gpsCoordinate1 ) )
              break;

            m_displayDriver.ClearScreenRange(DisplayDriver.DISPLAY_2, 3, 4, 0, 20);
            m_displayDriver.WriteMessage(DisplayDriver.DISPLAY_2, 3, 0,  m_globalVars.m_gpsCoordinate1.GetCoordinateName().toString() );
            m_displayDriver.WriteMessage(DisplayDriver.DISPLAY_2, 3, 8,  ":" );

            if ( m_gpsMath.CalculateDistance( lastCoordinate, m_globalVars.m_gpsCoordinate1, m_globalVars.m_fTemp11, GPSConvert.GPS_CONVERSION_NAUTICALMILES ) )
            {
              m_displayDriver.WriteMessage(DisplayDriver.DISPLAY_2, 3, 9, m_globalVars.m_fTemp11.floatFormat(52).toString() );

              if ( m_globalVars.m_fTemp11.compare(m_globalVars.m_fTemp12) < 0 )
                nWayPointSelected = n;
            }
          }

          m_displayDriver.ClearScreenRange(DisplayDriver.DISPLAY_2, 2, 3, 0, 20);
          m_displayDriver.WriteMessage(DisplayDriver.DISPLAY_2, 2, 0, "Selected waypoint:" );
          m_displayDriver.ClearScreenRange(DisplayDriver.DISPLAY_2, 3, 4, 0, 20);
          m_routeHandler.GetCoordinate( nRoute, nWayPointSelected, m_globalVars.m_gpsCoordinate1 );
          m_displayDriver.WriteMessage(DisplayDriver.DISPLAY_2, 3, 0,  m_globalVars.m_gpsCoordinate1.GetCoordinateName().toString() );
          CPU.delay(10000);

          return nWayPointSelected;
        }
      }



    return 0;
  }


  private boolean ChooseRoute(int nRoute) throws Exception
  {
    if ( !m_routeHandler.GetRouteName( nRoute, m_globalVars.m_buffer1 ))
      return false;

    m_displayDriver.ClearScreenRange(DisplayDriver.DISPLAY_1, 1, 2, 0, 20);
    m_displayDriver.WriteMessage(DisplayDriver.DISPLAY_1, 1, 0, m_globalVars.m_buffer1.toString());

    // Expect the number...
    MEMKeyInput input = m_keyboard.GetChar( true,
                                            MEMKey.KEYBOARD_MODE_SPECIAL,
                                            m_displayDriver,
                                            DisplayDriver.DISPLAY_1,
                                            0,
                                            0,
                                            m_globalVars.m_buffer1,
                                            0,
                                            MEMKey.KEYBOARD_BYTE_C,
                                            MEMKey.KEYBOARD_BYTE_D,
                                            MEMKey.KEYBOARD_NONE_BYTE,
                                            MEMKey.KEYBOARD_NONE_BYTE,
                                            MEMKey.KEYBOARD_NONE_BYTE );

      if (input.m_Key == MEMKey.KEYBOARD_BYTE_D)
        return true;

    return false;
  }

  private void DeleteRoute() throws Exception
  {
    if ( !AnyRoutes() ) return;

    m_displayDriver.WriteMessage(DisplayDriver.DISPLAY_1, 0, 0, "Delete route?");

    for ( int i = 0; i < 36; i++)
    {
      if (!m_routeHandler.IsRoute(i+1)) continue;

      if (ChooseRoute(i+1))
        m_routeHandler.DeleteRoute(i+1);
    }
  }
  private void HandleCreateRoute() throws Exception
  {
    int nRoute = m_routeHandler.GetNextAvailableRouteSlot();
    m_globalVars.m_buffer1.clear();
    m_globalVars.m_buffer1.append(nRoute);

    m_displayDriver.WriteMessage(DisplayDriver.DISPLAY_1, 0, 0, "Create route: ");
    m_displayDriver.WriteMessage(DisplayDriver.DISPLAY_1, 0, 13, m_globalVars.m_buffer1.toString() );

    m_displayDriver.WriteMessage(DisplayDriver.DISPLAY_2, 3, 0, "D=OK");
    m_displayDriver.WriteMessage(DisplayDriver.DISPLAY_2, 3, 12, "*=Cancel");

    m_displayDriver.WriteMessage(DisplayDriver.DISPLAY_1, 1, 0, "Name:");

    MEMKeyInput input = m_keyboard.GetChar( true,
                                            MEMKey.KEYBOARD_MODE_ENTER_TEXT,
                                            m_displayDriver,
                                            DisplayDriver.DISPLAY_1,
                                            2,
                                            0,
                                            m_globalVars.m_buffer1,
                                            20,
                                            MEMKey.KEYBOARD_BYTE_D,
                                            MEMKey.KEYBOARD_BYTE_STAR,
                                            MEMKey.KEYBOARD_BYTE_A,
                                            MEMKey.KEYBOARD_BYTE_B,
                                            MEMKey.KEYBOARD_BYTE_C );


    if (input.m_Key == MEMKey.KEYBOARD_BYTE_STAR ) return;
    m_routeHandler.SetRouteName( nRoute, m_globalVars.m_buffer1);

    // Add all the waypoints
    while ( true )
    {
      m_displayDriver.ClearScreenRange(DisplayDriver.DISPLAY_1, 3, 4, 0, 20);
      m_displayDriver.ClearScreenRange(DisplayDriver.DISPLAY_2, 0, 3, 0, 20);
      m_displayDriver.ShowCursor(DisplayDriver.DISPLAY_1|DisplayDriver.DISPLAY_2, false);

      int nWaypoint = m_routeHandler.GetNumberOfCoordinatesInRoute(nRoute)+1;
      if ( !CreateWaypoint(nWaypoint) ) break;

        // Store the waypoint
        if ( m_routeHandler.AddCoordinate(nRoute, m_globalVars.m_gpsCoordinate1) == 0 )
        {
          m_displayDriver.WriteMessage(DisplayDriver.DISPLAY_2, 2, 0, "*** EEPROM Error! ****");
          CPU.delay(3000);
        }
    }
  }

  private boolean CreateWaypoint(int nWaypoint)
  {

    m_globalVars.m_buffer1.clear();
    m_globalVars.m_buffer1.append("Waypoint:");
    m_globalVars.m_buffer1.append(nWaypoint);

    // Get the waypoint name
    m_displayDriver.WriteMessage(DisplayDriver.DISPLAY_1, 3, 0, m_globalVars.m_buffer1.toString());
    MEMKeyInput input = m_keyboard.GetChar( true,
                                            MEMKey.KEYBOARD_MODE_ENTER_TEXT,
                                            m_displayDriver, DisplayDriver.DISPLAY_1,
                                            3,
                                            11,
                                            m_globalVars.m_buffer1,
                                            8,
                                            MEMKey.KEYBOARD_BYTE_D,
                                            MEMKey.KEYBOARD_BYTE_STAR,
                                            MEMKey.KEYBOARD_BYTE_A,
                                            MEMKey.KEYBOARD_BYTE_B,
                                            MEMKey.KEYBOARD_BYTE_C );

    if (input.m_Key == MEMKey.KEYBOARD_BYTE_STAR) return (false);

    // Save the coordinate name
    m_globalVars.m_gpsCoordinate1.SetCoordinateName(m_globalVars.m_buffer1.toString());

    if ( !EditGPSCoordinate(DisplayDriver.DISPLAY_2, 0, m_globalVars.m_gpsCoordinate1 ) )
      return (false);

    return (true);
  }

  private void ComposeCoordinateString(int nDeg, int nMinute, int nMinuteDec )
  {
    m_globalVars.m_buffer1.clear();
    // Latitude/Longitude degrees
    if (nDeg<10)
      m_globalVars.m_buffer1.append("  ");
    else if (nDeg < 100)
      m_globalVars.m_buffer1.append(" ");

    m_globalVars.m_buffer1.append(nDeg);
    m_globalVars.m_buffer1.append((char)DisplayDriver.SPECIAL_CHARACTER_DEGREE);

    // Latitude/Longitude minutes
    if (nMinute<10)
      m_globalVars.m_buffer1.append(" ");

    m_globalVars.m_buffer1.append(nMinute);

    // Latitude/Longitude minute decimals
    m_globalVars.m_buffer1.append("'");
    m_globalVars.m_buffer1.append(nMinuteDec);
  }
  private boolean GetCoordinateFromDisplay(int nDisplay, int nRow, boolean bLongitude)
  {

    m_displayDriver.Goto(nDisplay, nRow, 6);

    //****************************************************//
    // Get the latitude/longitude (deg)
    //****************************************************//
    MEMKeyInput input = InternalGetNumber();
    if (input.m_Key == MEMKey.KEYBOARD_BYTE_STAR)
      return false;
    m_displayDriver.WriteCharacter(nDisplay, nRow, 6, (byte)input.m_Alfa);


    m_displayDriver.Goto(nDisplay, nRow, 7);
    input = InternalGetNumber();
    if (input.m_Key == MEMKey.KEYBOARD_BYTE_STAR)
      return false;
    m_displayDriver.WriteCharacter(nDisplay, nRow, 7, (byte)input.m_Alfa);


    m_displayDriver.Goto(nDisplay, nRow, 8);
    input = InternalGetNumber();
    if (input.m_Key == MEMKey.KEYBOARD_BYTE_STAR)
      return false;
    m_displayDriver.WriteCharacter(nDisplay, nRow, 8, (byte)input.m_Alfa);

    //****************************************************//
    // Get the latitude/longitude (min)
    //****************************************************//
    m_displayDriver.Goto(nDisplay, nRow, 10);
    input = InternalGetNumber();
    if (input.m_Key == MEMKey.KEYBOARD_BYTE_STAR)
      return false;
    m_displayDriver.WriteCharacter(nDisplay, nRow, 10, (byte)input.m_Alfa);

    m_displayDriver.Goto(nDisplay, nRow, 11);
    input = InternalGetNumber();
    if (input.m_Key == MEMKey.KEYBOARD_BYTE_STAR)
      return false;
    m_displayDriver.WriteCharacter(nDisplay, nRow, 11, (byte)input.m_Alfa);

    //****************************************************//
    // Get the latitude/longitude (min degrees)
    //****************************************************//
    m_displayDriver.Goto(nDisplay, nRow, 13);
    input = InternalGetNumber();
    if (input.m_Key == MEMKey.KEYBOARD_BYTE_STAR)
      return false;
    m_displayDriver.WriteCharacter(nDisplay, nRow, 13, (byte)input.m_Alfa);

    m_displayDriver.Goto(nDisplay, nRow, 14);
    input = InternalGetNumber();
    if (input.m_Key == MEMKey.KEYBOARD_BYTE_STAR)
      return false;
    m_displayDriver.WriteCharacter(nDisplay, nRow, 14, (byte)input.m_Alfa);

    m_displayDriver.Goto(nDisplay, nRow, 15);
    input = InternalGetNumber();
    if (input.m_Key == MEMKey.KEYBOARD_BYTE_STAR)
      return false;
    m_displayDriver.WriteCharacter(nDisplay, nRow, 15, (byte)input.m_Alfa);

    m_displayDriver.Goto(nDisplay, nRow, 16);
    input = InternalGetNumber();
    if (input.m_Key == MEMKey.KEYBOARD_BYTE_STAR)
      return false;
    m_displayDriver.WriteCharacter(nDisplay, nRow, 16, (byte)input.m_Alfa);


    m_displayDriver.Goto(nDisplay, nRow, 18);
    input = InternalGetNumber();
    if (input.m_Key == MEMKey.KEYBOARD_BYTE_STAR)
      return false;

    if (bLongitude)
      m_displayDriver.WriteCharacter(nDisplay, nRow, 18, input.m_Numeric == 3 ? (byte)'E' : (byte)'W');
    else
      m_displayDriver.WriteCharacter(nDisplay, nRow, 18, input.m_Numeric == 6 ? (byte)'N' : (byte)'S');


    return true;
  }

  public boolean EditGPSCoordinate(int nDisplay, int nRow, GPSCoordinate gpsCoordinate)
  {
    // Clear the 2 rows (we always have 2 rows because we need to enter the LAT & LONG)
    m_displayDriver.ClearScreenRange(nDisplay, nRow, (nRow+2), 0, 20);


    /********** Fix the latitude ************/
    {
      m_displayDriver.WriteMessage(nDisplay, nRow, 0, "LAT:");
      ComposeCoordinateString( gpsCoordinate.m_nLatDeg, gpsCoordinate.m_nLatMinute, gpsCoordinate.m_nLatMinuteDec);
      m_displayDriver.WriteMessage(nDisplay, nRow, 6, m_globalVars.m_buffer1.toString() );
      m_displayDriver.WriteCharacter(nDisplay, nRow, 18, (byte)(gpsCoordinate.m_bNorth ? 'N' : 'S'));
    }

    /********** Fix the longitude ************/
    {
      m_displayDriver.WriteMessage(nDisplay, (nRow+1), 0, "LONG:");
      ComposeCoordinateString( gpsCoordinate.m_nLongDeg, gpsCoordinate.m_nLongMinute, gpsCoordinate.m_nLongMinuteDec);
      m_displayDriver.WriteMessage(nDisplay, (nRow+1), 6, m_globalVars.m_buffer1.toString() );
      m_displayDriver.WriteCharacter(nDisplay, (nRow+1), 18, (byte)(gpsCoordinate.m_bEast ? 'E' : 'W'));
    }

    // Let the user enter the data
    m_displayDriver.ShowCursor(nDisplay, true);

    /********** Get the latitude ************/
    // Get the LAT
    if (!GetCoordinateFromDisplay(nDisplay, nRow, false ) )
      return false;

    int[] Row = m_displayDriver.GetRow(nDisplay, nRow);

    // Get the latitude deg
    m_globalVars.m_buffer1.clear();
    for (int nCol = 6; nCol < 9; nCol++)
      m_globalVars.m_buffer1.append( (char)(Row[nCol] & 0x00FF) );
    gpsCoordinate.m_nLatDeg = Integer.parseInt(m_globalVars.m_buffer1);

    // Get the latitude min
    m_globalVars.m_buffer1.clear();
    for (int nCol = 10; nCol < 12; nCol++)
      m_globalVars.m_buffer1.append( (char)(Row[nCol] & 0x00FF) );
    gpsCoordinate.m_nLatMinute = Integer.parseInt(m_globalVars.m_buffer1);

    // Get the latitude min deg
    m_globalVars.m_buffer1.clear();
    for (int nCol = 13; nCol < 17; nCol++)
      m_globalVars.m_buffer1.append( (char)(Row[nCol] & 0x00FF) );
    gpsCoordinate.m_nLatMinuteDec = Integer.parseInt(m_globalVars.m_buffer1);

    gpsCoordinate.m_bNorth = ((char)(Row[18] & 0x00FF) == 'N');

    /********** Get the longitude ************/
    // Get the LONG
    if (!GetCoordinateFromDisplay(nDisplay, nRow+1, true ) )
      return false;

    Row = m_displayDriver.GetRow(nDisplay, nRow+1);

    // Get the longitude deg
    m_globalVars.m_buffer1.clear();
    for (int nCol = 6; nCol < 9; nCol++)
      m_globalVars.m_buffer1.append( (char)(Row[nCol] & 0x00FF) );
    gpsCoordinate.m_nLongDeg = Integer.parseInt(m_globalVars.m_buffer1);

    // Get the latitude min
    m_globalVars.m_buffer1.clear();
    for (int nCol = 10; nCol < 12; nCol++)
      m_globalVars.m_buffer1.append( (char)(Row[nCol] & 0x00FF) );
    gpsCoordinate.m_nLongMinute = Integer.parseInt(m_globalVars.m_buffer1);

    // Get the latitude min deg
    m_globalVars.m_buffer1.clear();
    for (int nCol = 13; nCol < 17; nCol++)
      m_globalVars.m_buffer1.append( (char)(Row[nCol] & 0x00FF) );
    gpsCoordinate.m_nLongMinuteDec = Integer.parseInt(m_globalVars.m_buffer1);

    gpsCoordinate.m_bEast = ((char)(Row[18] & 0x00FF) == 'E');
    m_displayDriver.ShowCursor(DisplayDriver.DISPLAY_1|DisplayDriver.DISPLAY_2, false);

    return true;
  }

  private MEMKeyInput InternalGetNumber()
  {
    return ( m_keyboard.GetChar( true,
                                  MEMKey.KEYBOARD_MODE_NUMERIC,
                                  m_displayDriver,
                                  0,
                                  0,
                                  0,
                                  m_globalVars.m_buffer1,
                                  0,
                                  MEMKey.KEYBOARD_BYTE_STAR,
                                  MEMKey.KEYBOARD_BYTE_STAR,
                                  MEMKey.KEYBOARD_NONE_BYTE,
                                  MEMKey.KEYBOARD_NONE_BYTE,
                                  MEMKey.KEYBOARD_NONE_BYTE ) );
  }
}