package stamp.peripheral.sensor.GPS.NMEA;
import stamp.peripheral.sensor.GPS.StringOperations.*; // Contains some string handling classes
import stamp.peripheral.sensor.GPS.Include.*;
import stamp.peripheral.sensor.GPS.DisplayDriver.*;     // Contains the LCD driver for the CrystalFontz
import com.micromegacorp.math.hardware.*;
/**
  PE.060124
  Example (signal acquired): $GPGGA,092204.999,4250.5589,S,14718.5084,E,1,04,24.4,19.7,M,,,,0000*1F

  Field               Example           Comments
  Sentence ID         $GPGGA
  UTC Time            092204.999        hhmmss.sss
  Latitude            4250.5589         ddmm.mmmm
  N/S Indicator       S                 N = North, S = South
  Longitude           14718.5084        dddmm.mmmm
  E/W Indicator       E                 E = East, W = West
  Position Fix        1                 0 = Invalid, 1 = Valid SPS, 2 = Valid DGPS, 3 = Valid PPS
  Satellites Used     04                Satellites being used (0-12)
  HDOP                24.4              Horizontal dilution of precision
  Altitude            19.7              Altitude in meters according to WGS-84 ellipsoid
  Altitude Units      M                 M = Meters
  Geoid Seperation                      Geoid seperation in meters according to WGS-84 ellipsoid
  Seperation Units                      M = Meters
  DGPS Age                              Age of DGPS data in seconds
  DGPS Station ID                       0000
  Checksum           *1F
  Terminator          CR/LF

  How coordinates are represented:
  +----------------------+--------------------+----------------------+----------------------+--------------------+----------------------+
  |        2 bytes       |     2 bytes        |      2 bytes         |        2 bytes       |     2 bytes        |      2 bytes         |
  +----------------------+--------------------+----------------------+----------------------+--------------------+----------------------+
      1 bit -> N/S          16 bits -> lat min   16 bits -> min dec     1 bit -> E/W          16 bits->long min    16 bits -> min dec
     15 bit -> lat angle                                                15 bit -> long angle
*/

public class GGA extends NMEABase
{
  // Parts of the "GPGGA" message we care about
  private static final int GGA_MESSAGE_ID = 0;
  private static final int GGA_UTC_TIME = 1;
  private static final int GGA_LATITUDE = 2;
  private static final int GGA_LATITUDE_INDICATOR = 3; // N/S
  private static final int GGA_LONGITUDE = 4;
  private static final int GGA_LONGITUDE_INDICATOR = 5; // E/W
  private static final int GGA_POSITION_FIX = 6;
  private static final int GGA_END_OF_MESSAGE = 15; // Usually this is the "*"
  private static final int GGA_EXIT_MESSAGE = 16;

  private GPSCoordinate m_gpsCoordinate;
  private StringBuffer m_buffer;
  private UTCTimeParser m_timeParser;

  public GGA(UTCTimeParser timeParser)
  {
    super();

    m_gpsCoordinate = new GPSCoordinate();
    m_timeParser = timeParser;
    m_buffer = new StringBuffer(16);
  }
  public void Clear()
  {
    m_gpsCoordinate.Clear();
    super.Clear();
  }

  // Get
  public GPSCoordinate GetGPSCoordinate() { return m_gpsCoordinate; }

  public void GetLongitude( StringBuffer buffer )
  {
    m_gpsCoordinate.GetLongitude(buffer);
  }
  public void GetLatitude( StringBuffer buffer )
  {
    m_gpsCoordinate.GetLatitude(buffer);
  }
  // $GPGGA,210006.00,5535.4468,N,01300.3073,E,1,03,3.99,00035,M,040,M,,*66
  public boolean Parse(StringBuffer strMessage, Float32 fTemp)
  {
    Clear(); // Clear the content in this class

    try
    {
      int nLength = strMessage.length();

      // Nothing to do if the message is empty
      if ( nLength == 0 )
        return (false);

      m_bIsValid = false;

      char[] szArray = strMessage.toString().toCharArray();   // The character array containing the message
      int nMachineState = GGA_MESSAGE_ID;
      int nPosOld = 0;

      // Parse the message
      for ( int nPos = 0; ((nPos < nLength) && (nMachineState != GGA_EXIT_MESSAGE)); nPos++ )
      {
        nPosOld = nPos;
        nPos = StringOperations.IndexOfNext( nPos, szArray, ',' );
        m_strTemp.clear();
        m_buffer.clear();

        switch( nMachineState )
        {
          // Nothing to do here - it's the "$GPGGA" message ID
          case GGA_MESSAGE_ID: nMachineState = GGA_UTC_TIME; break;
          case GGA_UTC_TIME: nMachineState = GGA_LATITUDE; break;
          case GGA_LATITUDE:
          {
            m_buffer.clear();
            strMessage.subString(nPosOld, nPos, m_buffer );

            // ddmm.mmmm
            if (m_buffer.length()>=9)
            {
              // The first 2 characters are degrees
              m_buffer.subString(0, 2, m_strTemp );
              m_gpsCoordinate.m_nLatDeg = Integer.parseInt( m_strTemp );

              // The second 2 characters are the minutes
              m_buffer.subString(2, 4, m_strTemp );
              m_gpsCoordinate.m_nLatMinute = Integer.parseInt( m_strTemp );

              // The we have the last 4 characters are minutes decimals
              m_buffer.subString(5, 9, m_strTemp );
              m_gpsCoordinate.m_nLatMinuteDec = Integer.parseInt(m_strTemp);
            }
            nMachineState = GGA_LATITUDE_INDICATOR;   // Goto next state
          }
          break;
          case GGA_LATITUDE_INDICATOR:
          {
            // If this is the North or the South

            strMessage.subString(nPosOld, nPos, m_buffer );
            if ( m_buffer.length()>0)
            {
              char ch = m_buffer.charAt(0);
              if ((ch=='N')||(ch=='n'))
                m_gpsCoordinate.m_bNorth = true;
            }
            nMachineState = GGA_LONGITUDE;   // Goto next state
          }
          break;
          case GGA_LONGITUDE:
          {
            m_buffer.clear();
            strMessage.subString(nPosOld, nPos, m_buffer );

            // ddmm.mmmm
            if (m_buffer.length()>=10)
            {
              // The first 3 characters are degrees
              m_buffer.subString(0, 3, m_strTemp );
              m_gpsCoordinate.m_nLongDeg = Integer.parseInt( m_strTemp );

              // The second 2 characters are the minutes
              m_buffer.subString(3, 5, m_strTemp );
              m_gpsCoordinate.m_nLongMinute = Integer.parseInt( m_strTemp );

              // The we have the last 4 characters are minutes decimals
              m_buffer.subString(6, 10, m_strTemp );
              m_gpsCoordinate.m_nLongMinuteDec = Integer.parseInt(m_strTemp);
            }

            nMachineState = GGA_LONGITUDE_INDICATOR;   // Goto next state
          }
          break;
          case GGA_LONGITUDE_INDICATOR:
          {
            // If this is the East or the West
            strMessage.subString(nPosOld, nPos, m_buffer );
            if ( m_buffer.length()>0)
            {
              char ch = m_buffer.charAt(0);
              if ((ch=='E')||(ch=='e'))
                m_gpsCoordinate.m_bEast = true;
            }
            nMachineState = GGA_POSITION_FIX;   // Goto next state
          }
          break;
          case GGA_POSITION_FIX:
          {
            strMessage.subString(nPosOld, nPos, m_strTemp );
            if ( m_strTemp.length()>0)
              m_bIsValid = ( Integer.parseInt( m_strTemp.toString() ) == 1 );

            nMachineState = GGA_END_OF_MESSAGE;   // Goto next state
          }
          break;
          case GGA_END_OF_MESSAGE:
          {
            m_strTemp.clear();
            //strMessage.subString(nPosOld, nLength, m_strTemp ); // Get the rest of the message
            //m_bIsValid = ( m_strTemp.indexOf("*") >= 0 );
            m_bIsValid = ( strMessage.indexOf("*") >= 0 );

            // Now, check if this string starts with "*". Only in that case we know we've reached the end of message
            nMachineState = GGA_EXIT_MESSAGE;
          }
          break;
        }
      }
    }
    catch( Exception ex )
    {
      m_bIsValid = false;
    }

    return (m_bIsValid);

  }

}