package stamp.peripheral.sensor.GPS.NMEA;
import stamp.peripheral.sensor.GPS.StringOperations.*; // Contains some string handling classes
import stamp.peripheral.sensor.GPS.DisplayDriver.*;     // Contains the LCD driver for the CrystalFontz
import com.micromegacorp.math.hardware.*;

/**
PE.060126
$GPVTG Sentence (Course over ground)
Example (signal not acquired): $GPVTG,,T,,M,,N,,K*4E
Example (signal acquired): $GPVTG,89.68,T,,M,0.00,N,0.0,K*5F

Field               Example             Comments
Sentence ID         $GPVTG
Course              89.68               Course in degrees
Reference           89.68               T = True heading
Course                                  Course in degrees
Reference           89.68               M = Magnetic heading
Speed               0.00                Horizontal speed
Units               N                   N = Knots
Speed               0.00                Horizontal speed
Units               K                   K = KM/h
Checksum            *5F
Terminator                              CR/LF

*/

public class VTG extends NMEABase
{

  private final static int VTG_STATE_ID = 1;
  private final static int VTG_TRUE_HEADING = 2;
  private final static int VTG_TRUE_HEADING_Unit = 3; // Unit
  private final static int VTG_MAGN_HEADING = 4;
  private final static int VTG_MAGN_HEADING_Unit = 5; // Unit
  private final static int VTG_SPEED_N  = 6;
  private final static int VTG_SPEED_N_Unit  = 7;
  private final static int VTG_SPEED_M = 8;
  private final static int VTG_SPEED_M_Unit = 9;
  private final static int VTG_END_OF_MESSAGE = 10;
  private final static int VTG_EXIT_MESSAGE = 11;

  private int m_nTrueCourse=0;        // The true course
  private int m_nMagneticCourse=0;    // The magnetic course
  public int m_nSpeedKnots=0;        // Speed in knots * 100
  public int m_nSpeedKm=0;           // Speed in kilometers/hour * 100

  public VTG()
  {
  }
  public void Clear()
  {
    m_nTrueCourse=0;        // The true course
    m_nMagneticCourse=0;    // The magnetic course
    m_nSpeedKnots=0;        // Speed in knots
    m_nSpeedKm=0;           // Speed in kilometers/hour
  }

  public void GetCOGTrue(StringBuffer buffer)
  {
    buffer.clear();
    if ( m_nTrueCourse < 10 )
      buffer.append("00");
    else if ( m_nTrueCourse < 100 )
      buffer.append("0");

    buffer.append(m_nTrueCourse);
    buffer.append((char)DisplayDriver.SPECIAL_CHARACTER_DEGREE);
  }
  //public int GetMagneticCourse() { return ( m_nMagneticCourse ); };
  public void GetSOGKnots(StringBuffer buffer, Float32 fTemp)
  {
    buffer.clear();
    fTemp.set(m_nSpeedKnots);
    fTemp.divide(100);
    buffer.append(fTemp.floatFormat( 31 ));
    buffer.append("kts");
  }
  //public int GetSpeedKm() { return ( m_nSpeedKm ); };

  // $GPVTG,89.68,T,,M,0.00,N,0.0,K*5F
  public boolean Parse( StringBuffer strMessage, Float32 fTemp )
  {
    Clear(); // Clear the content in this class

    try
    {
      int nLength = strMessage.length();

      // Nothing to do if the message is empty
      if ( nLength == 0 )
        return (false);

      m_bIsValid = false;

      char[] szArray = strMessage.toString().toCharArray();   // The character array containing the message
      int nMachineState = VTG_STATE_ID;
      int nPosOld = 0;

      // Parse the message
      for ( int nPos = 0; ((nPos < nLength) && (nMachineState != VTG_EXIT_MESSAGE)); nPos++ )
      {
        nPosOld = nPos;
        nPos = StringOperations.IndexOfNext( nPos, szArray, ',' );
        m_strTemp.clear();

        switch( nMachineState )
        {
          // Nothing to do here - it's the "$GPVTG" message ID
          case VTG_STATE_ID: nMachineState = VTG_TRUE_HEADING; break;
          case VTG_TRUE_HEADING:
          {
            strMessage.subString(nPosOld, nPos, m_strTemp );
            m_nTrueCourse = Integer.parseInt(m_strTemp);
            nMachineState = VTG_TRUE_HEADING_Unit; // Goto next state
          }
          break;
          case VTG_TRUE_HEADING_Unit: { nMachineState = VTG_MAGN_HEADING; } break;

          case VTG_MAGN_HEADING:
          {
            strMessage.subString(nPosOld, nPos, m_strTemp );
            m_nMagneticCourse = Integer.parseInt(m_strTemp);
            nMachineState = VTG_MAGN_HEADING_Unit; // Goto next state
          }
          break;

          case VTG_MAGN_HEADING_Unit: { nMachineState = VTG_SPEED_N; } break;

          case VTG_SPEED_N:
          {
            strMessage.subString(nPosOld, nPos, m_strTemp);
            fTemp.set( m_strTemp.toString() );
            fTemp.multiply( 100 );
            m_nSpeedKnots = fTemp.intValue();
            nMachineState = VTG_SPEED_N_Unit;
          }
          break;

          case VTG_SPEED_N_Unit: { nMachineState = VTG_SPEED_M; } break;

          case VTG_SPEED_M:
          {
            strMessage.subString(nPosOld, nPos, m_strTemp );
            fTemp.set( m_strTemp.toString() );
            m_nSpeedKm = fTemp.intValue();
            nMachineState = VTG_SPEED_M_Unit;
          }
          break;
          case VTG_SPEED_M_Unit: { nMachineState = VTG_END_OF_MESSAGE; } break;
          case VTG_END_OF_MESSAGE:
          {
            //strMessage.subString(nPosOld, nLength, m_strTemp );

            // Now, check if this string starts with "*". Only in that case we know we've reached the end of message
            //m_bIsValid = ( m_strTemp.indexOf("*") >= 0 );
            m_bIsValid = ( strMessage.indexOf("*") >= 0 );

            nMachineState = VTG_EXIT_MESSAGE;
          }
          break;
        }
      }
    }
    catch( Exception ex )
    {
      m_bIsValid = false;
    }

    return (m_bIsValid);
  }
}