
package stamp.peripheral.sensor.GPS.NMEA;
import stamp.peripheral.sensor.GPS.StringOperations.*; // Contains some string handling classes
import stamp.peripheral.sensor.GPS.Include.*;

/**
  PE.060124
  Example (signal acquired): $GPZDA,024611.08,25,03,2002,00,00*6A

  UTC Time            024611.08   UTC time
  UTC Day             25          UTC day (01 to 31)
  UTC Month           03          UTC day (01 to 12)
  UTC Year            2002        UTC year (4 digit format)
  Local zone hours    00          Offset to local time zone in hours (+/- 00 to +/- 59)
  Local zone minutes  00          Offset to local time zone in minutes (00 to 59)
  Checksum            *6A
  Terminator        CR/LF
*/

public class ZDA extends NMEABase
{

  private final static int ZDA_STATE_ID = 1;
  private final static int ZDA_UTC_TIME = 2;
  private final static int ZDA_UTC_DAY = 3;
  private final static int ZDA_UTC_MONTH = 4;
  private final static int ZDA_UTC_YEAR = 5;
  private final static int ZDA_OFFSET_HOURS = 6;        // Not used in IQ Lassen
  private final static int ZDA_OFFSET_MINUTES = 7;      // Not used in IQ Lassen
  private final static int ZDA_END_OF_MESSAGE = 10;
  private final static int ZDA_EXIT_MESSAGE = 11;

  // Variables storing the UTC time
  private byte m_byteUTCHour = 0x00;      // Hour
  private byte m_byteUTCMinute = 0x00;    // Minute
  private byte m_byteUTCSecond = 0x00;    // Second

  private int m_nUTCDay;
  private int m_nUTCMonth;
  private int m_nUTCYear;
  private UTCTimeParser m_timeParser;
  private StringBuffer m_buffer;
  public boolean m_bIsValid;


  public ZDA(UTCTimeParser timeParser)
  {
    m_timeParser = timeParser;
    m_buffer = new StringBuffer(16);
  }
  public void Clear()
  {
    m_byteUTCHour=0;      // Hour
    m_byteUTCMinute=0;    // Minute
    m_byteUTCSecond=0;    // Second

    m_nUTCDay=0;
    m_nUTCMonth=0;
    m_nUTCYear=0;
    m_bIsValid = false;
    m_strTemp.clear();
  }
  public String GetID() { return ("$GPZDA"); };

  public void GetUTC(StringBuffer buffer)
  {
    UTCTimeParser.GetUTC(buffer, m_byteUTCHour, m_byteUTCMinute, m_byteUTCSecond);
  }

  public void GetDate(StringBuffer buffer )
  {
    UTCTimeParser.GetDate( buffer, m_nUTCMonth, m_nUTCDay, m_nUTCYear );
  }

  public int GetUTCDay(){ return m_nUTCDay; }
  public int GetUTCMonth(){ return m_nUTCMonth; }
  public int GetUTCYear(){ return m_nUTCYear; }

  // $GPZDA,024611.08,25,03,2002,00,00*6A
  public boolean Parse( StringBuffer strMessage )
  {
    Clear(); // Clear the content in this class

    try
    {
      int nLength = strMessage.length();

      // Nothing to do if the message is empty
      if ( nLength == 0 )
        return (false);

      m_bIsValid = false;

      char[] szArray = strMessage.toString().toCharArray();   // The character array containing the message
      int nMachineState = ZDA_STATE_ID;
      int nPosOld = 0;

      // Parse the message
      for ( int nPos = 0; ((nPos < nLength) && (nMachineState != ZDA_EXIT_MESSAGE)); nPos++ )
      {
        nPosOld = nPos;
        nPos = StringOperations.IndexOfNext( nPos, szArray, ',' );
        m_strTemp.clear();
        m_buffer.clear();

        switch( nMachineState )
        {
          // Nothing to do here - it's the "$GPZDA" message ID
          case ZDA_STATE_ID: nMachineState = ZDA_UTC_TIME; break;
          case ZDA_UTC_TIME:
          {
            strMessage.subString(nPosOld, nPos, m_buffer );
            if (m_timeParser.ParseUTCTime(m_buffer, m_strTemp))
            {
              m_byteUTCHour = m_timeParser.m_byteUTCHour;         // Hour
              m_byteUTCMinute = m_timeParser.m_byteUTCMinute;     // Minute
              m_byteUTCSecond = m_timeParser.m_byteUTCSecond;     // Second
            }
            nMachineState = ZDA_UTC_DAY; // Goto next state
          }
          break;
          case ZDA_UTC_DAY:
          {
            strMessage.subString(nPosOld, nPos, m_strTemp );
            if (m_strTemp.length()>0)
              m_nUTCDay = Integer.parseInt(m_strTemp);

            nMachineState = ZDA_UTC_MONTH;
          }
          break;
          case ZDA_UTC_MONTH:
          {
            strMessage.subString(nPosOld, nPos, m_strTemp );
            if (m_strTemp.length()>0)
              m_nUTCMonth = Integer.parseInt(m_strTemp);

            nMachineState = ZDA_UTC_YEAR;
          }
          break;

          case ZDA_UTC_YEAR:
          {
            strMessage.subString(nPosOld, nPos, m_strTemp );
            if (m_strTemp.length()>0)
              m_nUTCYear = Integer.parseInt(m_strTemp);
              if ( m_nUTCYear > 0 )
               m_nUTCYear -= 2000; // To get "06"

            nMachineState = ZDA_OFFSET_HOURS;
          }
          break;
          case ZDA_OFFSET_HOURS:
          {
            //strMessage.subString(nPosOld, nPos, m_strTemp );
            nMachineState = ZDA_END_OF_MESSAGE;
          }
          break;

          case ZDA_END_OF_MESSAGE:
          {
            //strMessage.subString(nPosOld, nLength, m_strTemp );

            // Now, check if this string starts with "*". Only in that case we know we've reached the end of message
            //m_bIsValid = ( m_strTemp.indexOf("*") >= 0 );
            m_bIsValid = ( strMessage.indexOf("*") >= 0 );
            m_bIsValid &= ( m_nUTCMonth > 0 );
            m_bIsValid &= ( m_nUTCYear > 0 );
            m_bIsValid &= ( m_nUTCDay > 0 );

            nMachineState=ZDA_EXIT_MESSAGE;
          }
          break;

        }

      }
    }
    catch( Exception ex )
    {
      m_bIsValid = false;
    }

    return (m_bIsValid);
  }
}