
package stamp.peripheral.sensor.GPS.RouteFollower;

import stamp.core.*;
import stamp.peripheral.memory.eeprom.*;                // Contains handling for EEPROM operations
import stamp.peripheral.sensor.GPS.GPSMath.*;           // Contains all the math calculations
import stamp.peripheral.sensor.GPS.DisplayDriver.*;     // Contains the LCD driver for the CrystalFontz
import stamp.peripheral.sensor.GPS.Include.*;           // Contains the linked list class and other stuff
import com.micromegacorp.math.hardware.*;               // Contains the uM-FPU implementation
import stamp.peripheral.sensor.GPS.NMEA.*;

public class RouteFollower
{
  public static final int FOLLOW_NONE = 0;          // We're not following any route
  public static final int FOLLOW_ASCENDING = 1;     // We're following the route ascendingly, ie. from 1->n
  public static final int FOLLOW_DESCENDING = 2;    // We're following the route descendingly ie. from n->1


  private int m_nRoute;      // The route we joined (current route)
  private int m_nWayPoint;   // Next waypoint (the waypoint number is the ID of the waypoint in the Route table 24LC256)
  private MemoryRouteHandler m_routeHandler;
  private int m_nFollowMode;
  GPSCoordinate m_gpsCoordinate;
  Settings m_settings;
  StringBuffer m_routeName;
  private boolean m_bIsMobState;

  public RouteFollower( MemoryRouteHandler routeHandler )
  {
    m_nFollowMode = FOLLOW_NONE;
    m_routeHandler = routeHandler;
    m_gpsCoordinate = new GPSCoordinate();  // Next waypoint
    m_routeName = new StringBuffer(25);
    m_settings = null;
    m_bIsMobState = false;
  }
  public void ToggleMobState(GPSCoordinate gpsCoordinate, DisplayDriver driver)
  {
    driver.ClearScreenRange(DisplayDriver.DISPLAY_2, 0, 3, 0, 20 );
    m_bIsMobState =!m_bIsMobState;
    if ( m_bIsMobState )
    {
      m_gpsCoordinate.Copy( gpsCoordinate );
      HandleDisconnectFromRoute( driver );
      return;
    }
  };
  public boolean IsMobState(){ return m_bIsMobState; }
  public void SetSettings(Settings settings){ m_settings = settings;}
  public void HandleDisconnectFromRoute(DisplayDriver driver)
  {
    m_nRoute = 0;
    m_nWayPoint = 0;
    m_nFollowMode = FOLLOW_NONE;

    // Save the settings to eeprom
    SaveSettings();

    driver.WriteMessage(DisplayDriver.DISPLAY_2, 0, 0, "Disconnected!");
    CPU.delay(10000);
    driver.ClearScreenRange(DisplayDriver.DISPLAY_2, 0, 3, 0, 20 );
  }
  public void HandleGotoNextWaypoint(DisplayDriver driver)
  {
    if (IsMobState()) return;
    int nNextWaypoint = m_nWayPoint;

    if (m_nFollowMode == FOLLOW_ASCENDING)
      nNextWaypoint++;
    else if (m_nFollowMode == FOLLOW_DESCENDING)
      nNextWaypoint--;;

    JoinRoute( m_nRoute, nNextWaypoint, m_nFollowMode, driver);
  }
  public void HandleGotoPreviousWaypoint(DisplayDriver driver)
  {
    if (IsMobState()) return;
    int nNextWaypoint = m_nWayPoint;

    if (m_nFollowMode == FOLLOW_ASCENDING)
      nNextWaypoint--;
    else if (m_nFollowMode == FOLLOW_DESCENDING)
      nNextWaypoint++;;

    JoinRoute( m_nRoute, nNextWaypoint, m_nFollowMode, driver);
  }

  public void SetStoredSettings()
  {
    m_nRoute = m_settings.m_nRoute;
    m_nWayPoint = m_settings.m_nWayPoint;
    m_nFollowMode = m_settings.m_nFollowMode;
    if (!ResolveWaypoint( m_nRoute, m_nWayPoint))
    {
      m_nRoute = 0;
      m_nWayPoint = 0;
      m_nFollowMode = FOLLOW_NONE;
    }
  }

  private void SaveSettings()
  {
    m_settings.m_nRoute = m_nRoute;
    m_settings.m_nWayPoint = m_nWayPoint;
    m_settings.m_nFollowMode = m_nFollowMode;
    m_settings.SaveSettings( m_routeHandler.GeteepromMemory() );
  }

  public boolean JoinRoute( int nRoute, int nWayPoint, int nFollowMode, DisplayDriver driver)
  {
    // Try to get the waypoint from memory
    if (!ResolveWaypoint(nRoute, nWayPoint))
      return (false);

    // Store the variables
    m_nRoute = nRoute;
    m_nWayPoint = nWayPoint;
    m_nFollowMode = nFollowMode;

    // Save the route information to EEPROM
    SaveSettings();

    driver.WriteMessage(DisplayDriver.DISPLAY_2, 0, 0, "Route updated!");
    CPU.delay(10000);
    driver.ClearScreenRange(DisplayDriver.DISPLAY_2, 0, 3, 0, 20 );

    return (true);
  }
  private boolean ResolveWaypoint(int nRoute, int nWayPoint)
  {
    try
    {
      if (!m_routeHandler.GetCoordinate(nRoute, nWayPoint, m_gpsCoordinate ) )
        return (false);

      m_routeName.clear();
      if (!m_routeHandler.GetRouteName( nRoute, m_routeName ) )
        return (false);
    }
    catch(Exception ex)
    {
      // Nothing to do here
    }
    return (true);
  }
  //
  public boolean Update( GPSCoordinate coordinate,      // The current position reported by the GPS
                         GPSMath gpsMath,               // Class containing the Math calculations
                         DisplayDriver displayDriver,   // Class handling the display driver
                         GlobalVars globalVars,         // Class handling the global variables
                         NMEA0183 nmea0183 )            // Class handling the messages received from the IQ Lassen module
  {

    StringBuffer buffer = globalVars.m_buffer1;
    buffer.clear();


    // Is there any VTG message?
    VTG vtg = nmea0183.GetVTGMessage();
    if ( vtg != null )
    {
      displayDriver.WriteMessage(DisplayDriver.DISPLAY_1, 2, 0, "COG:" );

      vtg.GetCOGTrue( globalVars.m_buffer1 );
      displayDriver.WriteMessage(DisplayDriver.DISPLAY_1, 2, 4, globalVars.m_buffer1.toString() );

      displayDriver.WriteMessage(DisplayDriver.DISPLAY_1, 2, 9, "SOG:" );
      vtg.GetSOGKnots( globalVars.m_buffer1, globalVars.m_fTemp1 );
      displayDriver.WriteMessage(DisplayDriver.DISPLAY_1, 2, 13, globalVars.m_buffer1.toString() );
    }

    // Show the SMG
    if ( nmea0183.GetSpeedMadeGood( globalVars.m_fTemp1 ) )
    {
      buffer.clear();
      displayDriver.WriteMessage(DisplayDriver.DISPLAY_1, 3, 0, "SMG:" );
      buffer.append( globalVars.m_fTemp1.floatFormat( 41 ) );
      buffer.append( "kts" );
      displayDriver.WriteMessage(DisplayDriver.DISPLAY_1, 3, 4, buffer.toString() );
    }

    /*For debugging purposes...
    int nFreeMemory = Memory.freeMemory();
    buffer.clear();
    buffer.append(nFreeMemory);
    displayDriver.WriteMessage(DisplayDriver.DISPLAY_1, 3, 14, buffer.toString() );
    */

    if ( m_nFollowMode == FOLLOW_NONE && !IsMobState() ) return true;


    Float32 fCourse = globalVars.m_fTemp11;
    Float32 fDistance = globalVars.m_fTemp12;

    // Display the name of the next point
    displayDriver.WriteMessage(DisplayDriver.DISPLAY_2, 0, 0, IsMobState() ? "MOB" : m_gpsCoordinate.GetCoordinateName().toString() );

    // Next point coordinates
    buffer.clear();
    m_gpsCoordinate.GetLatitude( buffer );
    displayDriver.WriteMessage(DisplayDriver.DISPLAY_2, 0, 11, buffer.toString() );

    buffer.clear();
    m_gpsCoordinate.GetLongitude( buffer );
    displayDriver.WriteMessage(DisplayDriver.DISPLAY_2, 1, 11, buffer.toString() );

    if (!gpsMath.CalculateCourseAndDistance( coordinate, m_gpsCoordinate, fCourse, fDistance, GPSConvert.GPS_CONVERSION_DEGREES, GPSConvert.GPS_CONVERSION_NAUTICALMILES ))
    {
      displayDriver.WriteMessage(DisplayDriver.DISPLAY_2, 1, 0, "** GPSMath failed **");
      return (false);
    }

    // Write the bearing
    buffer.clear();
    int nCourse = fCourse.intValue();
    if ( nCourse < 10 )
      buffer.append("00");
    else if ( nCourse < 100 )
      buffer.append("0");

    buffer.append(nCourse);
    buffer.append((char)DisplayDriver.SPECIAL_CHARACTER_DEGREE);
    displayDriver.WriteMessage(DisplayDriver.DISPLAY_2, 1, 0, "BTW:");
    displayDriver.WriteMessage(DisplayDriver.DISPLAY_2, 1, 4, buffer.toString());


    // Write the "DTG"
    displayDriver.WriteMessage(DisplayDriver.DISPLAY_2, 2, 0, "DTG:");
    buffer.clear();

    buffer.append(fDistance.floatFormat(63).toString());
    buffer.append("nm");
    displayDriver.WriteMessage(DisplayDriver.DISPLAY_2, 2, 4, buffer.toString());

    // Clear the ETA
    displayDriver.ClearScreenRange(DisplayDriver.DISPLAY_2, 2, 3, 12, 20 );

    // Calculate the ETA
    if ( vtg != null )
    {
      if ( vtg.m_nSpeedKnots > 0 )
      {
        nmea0183.GetSpeedMadeGood( globalVars.m_fTemp1 );
        fDistance.divide(globalVars.m_fTemp1);
        displayDriver.WriteMessage(DisplayDriver.DISPLAY_2, 2, 12, "ETA:");

        displayDriver.WriteMessage(DisplayDriver.DISPLAY_2, 2, 16, fDistance.floatFormat(31).toString());
        displayDriver.WriteMessage(DisplayDriver.DISPLAY_2, 2, 19, "h");
      }
    }

    return (false);
  }

}