package stamp.peripheral.sensor.humidity.sht11;
import stamp.core.*;
import stamp.math.*;

/*
  STAMP                 SENSIRION
  p1 ----o-----/\/\---o--pin 3 sck, clock
                      |
         ;---/\/\-----' pull-down
         | 4.7k
  Vss--o-o---------------pin 1 common
       |
      ===  0.1uf
       |
  Vdd--o-o---------------pin 4 +5 volts
         |    4.7k
         '---/\/\-----; pull-up
                      |
               220    |
  p0 ----------/\/\---o--pin 2 dta, data
*/

public class sht11 {

  private static final int shtTemp = 0x0003;
  private static final int shtHumi = 0x0005;
  private static final int shtStatW = 0x0006;
  private static final int shtStatR = 0x0007;
  private static final int shtReset = 0x001E;

  private int datapin,clockpin;

  public sht11(int datapin, int clockpin) {
    this.datapin = datapin;
    this.clockpin = clockpin;
    connectionReset();
  }

  // This function makes the SHT11 is ready with the measurement or ~1/4 seconds. According to the
  // SHT11 documentation, the measurement can take up to 210 ms
  private void waitForMeasurement() {
    CPU.writePin(datapin,true);
    CPU.setInput(datapin);
    for (int i=0; i<125; i++) { //wait 125*2 msec
      if (!CPU.readPin(datapin)) break;
      CPU.delay(21); //2 msec delay
    }
  }

  // connection reset: 9 clock cyles with ShtData high, then start sequence
  // Function initializes the SHT11 chip
  // generates SHT11 "connection reset" sequence
  //           ______________________________________________
  // ShtData
  //              ___     ___     ___     ____     ___     __.
  // Clock    ___| 1 |___| 2 |___| 3 |___|4-8 |___| 9 |___|
  public void connectionReset() {
    CPU.writePin(datapin,true);
    CPU.setInput(datapin);
    CPU.writePin(clockpin,false);
    for (int i=0; i<9; i++) {
      CPU.writePin(clockpin,true);
      CPU.writePin(clockpin,false);
    }
    CPU.writePin(clockpin,true);
  }

  //                         ___
  // ShtData        |_______|
  //                _     _
  // Clock           |___|
  // generates SHT11 "start" sequence
  private void transmissionStart() {
    // ensure we have the correct values
    CPU.writePin(datapin,true);
    CPU.writePin(clockpin,true);
    CPU.writePin(datapin,false);
    CPU.writePin(clockpin,false);
    CPU.writePin(clockpin,true);
    CPU.writePin(datapin,true);
    CPU.writePin(clockpin,false);
    CPU.setInput(datapin);
  }

  public int rawHumidity() {
    transmissionStart(); // Alert device
    writeByte(shtHumi); // send humidity command
    waitForMeasurement();    // wait for measurement
    int humidity = readByte(true) << 8; // Get the MSB
    humidity |= readByte(false);
    return humidity;
  }

/*
  Some math explanation (as found at http://www.emesystems.com/OL2sht1x.htm)

  Here is the Sensirion formula that converts the raw result into relative humidity, a quadratic polynomial:

  RH = c1 + c2* result + c3 * result2
  with constants as follows:

  c1 = -4c2 = 0.0405c3 = -2.8E-6
  The formula can be rewritten as follows:

  RH = (c2 + (c3 * result)) * result + c1
  The Stamp translation is the following. Each term is going to be multiplied times 10,in order to carry through one digit to the right of the decimal point.:

  RH=(26542-(54723**result+result))**result - 40
      -----  --------------------  --------   --
     ( c2 + c3 * result ) *result + c1

  with constants as follows:

  10*c1 = -40 and on the stamp, that is the -40 at the end of the expression above
  10*c2 = 0.405 on the stamp written as the fraction 26542/65536, (denominator implied) or
          26542 ** which will be used with the ** result that comes after adding... (the tricky one!)
  (10*c3 * result) = -2.8E-5 * result 2.8E-5 * result
          is the same as the fraction 1.835/65536 * result
          is the same as 1.835 * result / 65536
          is the same as (0.835*result + result)/65536
          is the same as (54723/65536 * result + result) / 65536
          which on the Stamp is done with the denominator implied in the **. (54723 ** result + result) **

*/
  /*
  public int relHumidity(int rawHumidity) {
    int RH = rawHumidity + UnsignedIntMath.umulf(rawHumidity,(short)54723);
    return UnsignedIntMath.umulf(rawHumidity,26542-RH) - 40;
  }
  */

/*
  Similar logic (see explanation above) applies for the temperature compensation

  From Sensirion, it involves the deviation of the temperature from 25 degrees Celsius, as well as
  the raw humidity value (result), added to the uncompensated humidity value:

  RHtc = (T - 25) * (0.01 + 0.00008 * result) + RH
  Order of magnitude: The raw humidity variable "result" can have values from near zero up to about 3000.
  So the temperature compensation factor calculated by the expression inside the second set of parentheses
  varies from 0.01 percent to 0.25 percent per degree Celsius. Suppose the humidity is near 100 percent.
  Then the humidity compensation could add or subtract 6 percent humidity over the zero to 50 degree Celsius range.
  Less compensation at lower humidities.

  Negative values don't work with the Stamp integer math ** operator,
  so I transform the temperature to Kelvin, by adding 273.

  RHtc = ((T+273) - 25) * (0.01 + 0.00008 * result) - (273 * (0.01 + 0.00008 * result)) + RH
          ------------
             (T+248)

  That might seem more complicated, adding the 273 at one point and subtracting it at another,
  but the benefit is that the quantity (T+248) is always positive, and will multiply using the ** operator .

  Consider the factor that involves the raw humidity "result".

  (0.01 + 0.00008 * result)
  It takes a treament just like the one above.

  0.01 on the Stamp is written as 655/65536
  0.00008 * result
  is the same as 5.243/65536 * result which
  is the same as (5* result + 0.243* result)/65536 which
  is the same as (5*result + 15917/65536 * result)/65536
  which on the Stamp is done with the denominator 65536 implied in the ** operator and combining the two terms,
  the factor to be multiplied times temperature is:
  (655 + (5*result) + (15917 ** result)) ** (deviation of temperature from 25 Celsius)

  RHtc=655+(result*5)+(15917**result)
  For example, suppose result = 1742 when RH=58.0%. The factor is computed on a calculator as follows:
  (0.01 + 0.00008*1742) = (0.01+0.1394) = 0.1494. (that change in %RH per degree Celsius)
  And in Stampese,
  655 + (1742 * 5) + (1742 ** 15917) = 655 + 8710 + 423 = 655+9133 = 9788.
  Observe that 9788/65536 = .14935, which is, as it should be,
  the same as the number computed on a scientific calculator. This is the number required for the ** operator.

  The final formula is this:

  RHtc=(RHtc**(degC+2480))-(RHtc**2730)+RH ' compensated value
  Going in, the factor we just calculated is held temporarily in variable RHtc.
  The temperature is held in variable degC, but recall that the value is actually
  ten times temperature. That is okay, because raw humidity variable RH is also holdes a value
  that is ten times the real value. The absolute value constant is 2730 in units of tenths of a degree.
*/
  /*
  public int tcHumidity(int celsius, int rawHumidity, int relHumidity) {
    int RHtc = 655 + (rawHumidity*5) + UnsignedIntMath.umulf(rawHumidity,15917);
    return UnsignedIntMath.umulf(celsius+2480,RHtc) - UnsignedIntMath.umulf(2730,RHtc) + relHumidity;
  }
  */

  public int rawTemp() {
    transmissionStart();   // Alert device
    writeByte(shtTemp);  // send temperature command
    waitForMeasurement();     // wait for measurement
    int temp = readByte(true) << 8; // Get the MSB
    temp |= readByte(false);
    return temp;
  }

  public int celsius(int rawTemp) {
    return ((rawTemp+5)/10)-400; //from 100ths to 10ths of a degree with rounding
  }

  private int readByte(boolean bSendAck) {
    CPU.writePin(clockpin,false);   // Prepare for shiftIn command
    int nData = CPU.shiftIn(datapin,clockpin,8,CPU.PRE_CLOCK_MSB);
    CPU.writePin(clockpin,false);
    if (bSendAck) {
      // Send the acknowledge bit
      CPU.writePin(datapin,false);
      CPU.writePin(clockpin,false);
      CPU.writePin(clockpin,true);
      CPU.writePin(clockpin,false);
      CPU.setInput(datapin);
    }
    else {
      // Skip the acknowledge bit
      CPU.shiftOut(datapin,clockpin,1,CPU.SHIFT_LSB,1);
    }
    return nData;
  }

  private void writeByte(int data) {
    CPU.writePin(clockpin,false);
    data = data << 8;
    CPU.shiftOut(datapin,clockpin,8,CPU.SHIFT_MSB,data);
    CPU.writePin(clockpin,false);   // Prepare for shiftIn command
    CPU.shiftIn(datapin,clockpin,1,CPU.POST_CLOCK_LSB); // Get the acknowledge bit...
  }

}